import pytest

import datetime

from acondbs.db.sa import sa
from acondbs.models import (
    ProductType,
    FieldType,
    Field,
    TypeFieldAssociation,
    Product,
    AttributeUnicodeText,
    AttributeDate,
    AttributeDateTime,
)


##__________________________________________________________________||
@pytest.fixture
def app(app_empty):

    y = app_empty

    # create fields
    field1 = Field(name="attr1", type_=FieldType.UnicodeText)
    field2 = Field(name="date_produced", type_=FieldType.Date)
    field3 = Field(name="time_posted", type_=FieldType.DateTime)
    fields = [field1, field2, field3]

    # create product types
    Map = ProductType(
        type_id=1,
        name="map",
        order=2,
        indef_article="a",
        singular="map",
        plural="maps",
        icon="mdi-map",
        fields=[TypeFieldAssociation(field=f) for f in fields],
    )
    Beam = ProductType(
        type_id=2,
        name="beam",
        order=1,
        indef_article="a",
        singular="beam",
        plural="beams",
        icon="mdi-spotlight-beam",
        fields=[TypeFieldAssociation(field=f) for f in fields],
    )

    # create products
    map1 = Product(product_id=1, name="map1", type_=Map)
    AttributeUnicodeText(product=map1, field=field1, name="attr1", value="value1")
    AttributeDate(
        product=map1, field=field2, name="date_produced", value=datetime.date(2020, 2, 1)
    )
    AttributeDateTime(
        product=map1,
        field=field3,
        name="time_posted",
        value=datetime.datetime(2020, 2, 1, 9, 10, 25),
    )
    map2 = Product(product_id=2, name="map2", type_=Map)
    AttributeUnicodeText(product=map2, field=field1, name="attr1", value="value2")
    AttributeDate(
        product=map2, field=field2, name="date_produced", value=datetime.date(2020, 2, 10)
    )
    AttributeDateTime(
        product=map2,
        field=field3,
        name="time_posted",
        value=datetime.datetime(2020, 2, 10, 13, 20, 2),
    )
    map3 = Product(product_id=3, name="map3", type_=Map)
    AttributeUnicodeText(product=map3, field=field1, name="attr1", value="value3")
    AttributeDate(
        product=map3, field=field2, name="date_produced", value=datetime.date(2020, 3, 19)
    )
    AttributeDateTime(
        product=map3,
        field=field3,
        name="time_posted",
        value=datetime.datetime(2020, 3, 20, 8, 45, 30),
    )
    beam1 = Product(product_id=4, name="beam1", type_=Beam)
    AttributeUnicodeText(product=beam1, field=field1, name="attr1", value="value4")
    AttributeDate(
        product=beam1, field=field2, name="date_produced", value=datetime.date(2020, 2, 5)
    )
    AttributeDateTime(
        product=beam1,
        field=field3,
        name="time_posted",
        value=datetime.datetime(2020, 2, 5, 12, 3, 48),
    )
    beam2 = Product(product_id=5, name="beam2", type_=Beam)
    AttributeUnicodeText(product=beam2, field=field1, name="attr1", value="value5")
    AttributeDate(
        product=beam2, field=field2, name="date_produced", value=datetime.date(2020, 3, 4)
    )
    AttributeDateTime(
        product=beam2,
        field=field3,
        name="time_posted",
        value=datetime.datetime(2020, 3, 4, 19, 22, 5),
    )

    with y.app_context():
        sa.session.add(Map)
        sa.session.add(Beam)
        sa.session.commit()

    yield y


##__________________________________________________________________||



import graphene

from ...funcs import get_git_hub_viewer_from_info
from .. import type_

from .... import ops


##__________________________________________________________________||
def _reshape_arg_attributes(attributes):
    ret = {e["field_id"]: e["value"] for t, v in attributes.items() for e in v}
    return ret


##__________________________________________________________________||
class RelationInputFields(graphene.InputObjectType):
    """A relation to another product"""

    product_id = graphene.Int(
        required=True,
        description="The product ID of the other product",
    )
    type_id = graphene.Int(
        required=True,
        description="The relation type ID",
    )


class AttributeUnicodeTextInputFields(graphene.InputObjectType):
    field_id = graphene.Int(
        required=True,
        description="The field ID of the attribute",
    )
    value = graphene.String(
        description="The value of the attribute",
    )


class AttributeBooleanInputFields(graphene.InputObjectType):
    field_id = graphene.Int(
        required=True,
        description="The field ID of the attribute",
    )
    value = graphene.Boolean(
        description="The value of the attribute",
    )


class AttributeIntegerInputFields(graphene.InputObjectType):
    field_id = graphene.Int(
        required=True,
        description="The field ID of the attribute",
    )
    value = graphene.Int(
        description="The value of the attribute",
    )


class AttributeFloatInputFields(graphene.InputObjectType):
    field_id = graphene.Int(
        required=True,
        description="The field ID of the attribute",
    )
    value = graphene.Float(
        description="The value of the attribute",
    )


class AttributeDateInputFields(graphene.InputObjectType):
    field_id = graphene.Int(
        required=True,
        description="The field ID of the attribute",
    )
    value = graphene.Date(
        description="The value of the attribute",
    )


class AttributeDateTimeInputFields(graphene.InputObjectType):
    field_id = graphene.Int(
        required=True,
        description="The field ID of the attribute",
    )
    value = graphene.DateTime(
        description="The value of the attribute",
    )


class AttributeTimeInputFields(graphene.InputObjectType):
    field_id = graphene.Int(
        required=True,
        description="The field ID of the attribute",
    )
    value = graphene.Time(
        description="The value of the attribute",
    )


class AttributesInputFields(graphene.InputObjectType):
    unicode_text = graphene.InputField(
        graphene.List(AttributeUnicodeTextInputFields),
        description=("Attributes of type unicode text"),
    )
    boolean = graphene.InputField(
        graphene.List(AttributeBooleanInputFields),
        description=("Attributes of type boolean"),
    )
    integer = graphene.InputField(
        graphene.List(AttributeIntegerInputFields),
        description=("Attributes of type integer"),
    )
    float = graphene.InputField(
        graphene.List(AttributeFloatInputFields),
        description=("Attributes of type float"),
    )
    date = graphene.InputField(
        graphene.List(AttributeDateInputFields),
        description=("Attributes of type date"),
    )
    date_time = graphene.InputField(
        graphene.List(AttributeDateTimeInputFields),
        description=("Attributes of type date time"),
    )
    time = graphene.InputField(
        graphene.List(AttributeTimeInputFields),
        description=("Attributes of type time"),
    )


class CommonInputFields:
    """Common input fields of mutations for creating and updating products"""

    contact = graphene.String(
        description=(
            "A person or group that can be contacted for questions or "
            "issues about the product."
        )
    )
    note = graphene.String(
        description="Note about the product in MarkDown.",
    )
    paths = graphene.List(
        graphene.String,
        description="Paths to the products. e.g., nersc:/go/to/my/product_v3",
    )
    relations = graphene.InputField(
        graphene.List(RelationInputFields),
        description=("Relations to other products"),
    )
    attributes = graphene.InputField(
        AttributesInputFields,
        description="Attributes",
    )


class CreateProductInput(graphene.InputObjectType, CommonInputFields):
    """Input to createProduct()"""

    type_id = graphene.Int(
        required=True,
        description="The product type ID",
    )
    name = graphene.String(
        required=True,
        description="The name of the product",
    )
    date_produced = graphene.Date(
        description="The date on which the product was produced"
    )
    produced_by = graphene.String(
        description="The person or group that produced the product"
    )
    posted_by = graphene.String(
        description="The person who entered the DB entry."
    )


class UpdateProductInput(graphene.InputObjectType, CommonInputFields):
    """Input to updateProduct()"""

    updated_by = graphene.String(
        description="The person who updated the DB entry."
    )


##__________________________________________________________________||
class CreateProduct(graphene.Mutation):
    """Create a product"""

    class Arguments:
        input = CreateProductInput(
            required=True,
            description="the input to createProduct()",
        )

    ok = graphene.Boolean()
    product = graphene.Field(lambda: type_.Product)

    def mutate(root, info, input):

        viewer = get_git_hub_viewer_from_info(info)
        posting_git_hub_user_id = viewer.user_id

        attributes = input.pop("attributes", {})
        attributes = _reshape_arg_attributes(attributes)

        model = ops.create_product(
            attributes=attributes,
            posting_git_hub_user_id=posting_git_hub_user_id,
            **input
        )

        ops.commit()
        ok = True
        return CreateProduct(product=model, ok=ok)


class UpdateProduct(graphene.Mutation):
    """Update a product.

    Note: This is to update the DB entry about a product. If the
    product itself has been updated, a new entry should be added by
    createProduct()
    """

    class Arguments:
        product_id = graphene.Int(
            required=True,
            description="The productId of a product to be updated.",
        )
        input = UpdateProductInput(
            required=True,
            description="an input to updateProduct()",
        )

    ok = graphene.Boolean()
    product = graphene.Field(lambda: type_.Product)

    def mutate(root, info, product_id, input):

        viewer = get_git_hub_viewer_from_info(info)
        updating_git_hub_user_id = viewer.user_id

        attributes = input.pop("attributes", {})
        attributes = _reshape_arg_attributes(attributes)

        model = ops.update_product(
            product_id=product_id,
            attributes=attributes,
            updating_git_hub_user_id=updating_git_hub_user_id,
            **input
        )
        ops.commit()
        ok = True
        return UpdateProduct(product=model, ok=ok)


class DeleteProduct(graphene.Mutation):
    """Delete a product"""

    class Arguments:
        product_id = graphene.Int(
            required=True,
            description="The productId of a product to be deleted.",
        )

    ok = graphene.Boolean()

    def mutate(root, info, product_id):
        ops.delete_product(product_id)
        ops.commit()
        ok = True
        return DeleteProduct(ok=ok)


##__________________________________________________________________||



import pytest

from acondbs import ops

@pytest.fixture
def app(app_users):
    y = app_users

    with y.app_context():
        ops.create_log(level="DEBUG", message="A debug message!")
        ops.create_log(level="ERROR", message="An error message!")
        ops.commit()

    yield y



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema_error[error-immutable-field] 1'] = {
    'data': {
        'allProductRelationTypes': {
            'edges': [
                {
                    'node': {
                        'indefArticle': 'a',
                        'name': 'parent',
                        'plural': 'parents',
                        'relations': {
                            'edges': [
                                {
                                    'node': {
                                        'other': {
                                            'name': 'map1',
                                            'productId': '1'
                                        },
                                        'self_': {
                                            'name': 'beam1',
                                            'productId': '4'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'other': {
                                            'name': 'map1',
                                            'productId': '1'
                                        },
                                        'self_': {
                                            'name': 'beam2',
                                            'productId': '5'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'other': {
                                            'name': 'beam1',
                                            'productId': '4'
                                        },
                                        'self_': {
                                            'name': 'beam2',
                                            'productId': '5'
                                        }
                                    }
                                }
                            ]
                        },
                        'reverse': {
                            'name': 'child',
                            'typeId': '2'
                        },
                        'singular': 'parent',
                        'typeId': '1'
                    }
                },
                {
                    'node': {
                        'indefArticle': 'a',
                        'name': 'child',
                        'plural': 'children',
                        'relations': {
                            'edges': [
                                {
                                    'node': {
                                        'other': {
                                            'name': 'beam1',
                                            'productId': '4'
                                        },
                                        'self_': {
                                            'name': 'map1',
                                            'productId': '1'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'other': {
                                            'name': 'beam2',
                                            'productId': '5'
                                        },
                                        'self_': {
                                            'name': 'map1',
                                            'productId': '1'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'other': {
                                            'name': 'beam2',
                                            'productId': '5'
                                        },
                                        'self_': {
                                            'name': 'beam1',
                                            'productId': '4'
                                        }
                                    }
                                }
                            ]
                        },
                        'reverse': {
                            'name': 'parent',
                            'typeId': '1'
                        },
                        'singular': 'child',
                        'typeId': '2'
                    }
                },
                {
                    'node': {
                        'indefArticle': 'a',
                        'name': 'plaintiff',
                        'plural': 'plaintiffs',
                        'relations': {
                            'edges': [
                            ]
                        },
                        'reverse': {
                            'name': 'defendant',
                            'typeId': '4'
                        },
                        'singular': 'plaintiff',
                        'typeId': '3'
                    }
                },
                {
                    'node': {
                        'indefArticle': 'a',
                        'name': 'defendant',
                        'plural': 'defendants',
                        'relations': {
                            'edges': [
                            ]
                        },
                        'reverse': {
                            'name': 'plaintiff',
                            'typeId': '3'
                        },
                        'singular': 'defendant',
                        'typeId': '4'
                    }
                }
            ]
        }
    }
}

snapshots['test_schema_success[update] 1'] = {
    'data': {
        'updateProductRelationType': {
            'ok': True,
            'productRelationType': {
                'indefArticle': 'an',
                'name': 'parent',
                'plural': 'mmaps',
                'relations': {
                    'edges': [
                        {
                            'node': {
                                'other': {
                                    'name': 'map1',
                                    'productId': '1'
                                },
                                'self_': {
                                    'name': 'beam1',
                                    'productId': '4'
                                }
                            }
                        },
                        {
                            'node': {
                                'other': {
                                    'name': 'map1',
                                    'productId': '1'
                                },
                                'self_': {
                                    'name': 'beam2',
                                    'productId': '5'
                                }
                            }
                        },
                        {
                            'node': {
                                'other': {
                                    'name': 'beam1',
                                    'productId': '4'
                                },
                                'self_': {
                                    'name': 'beam2',
                                    'productId': '5'
                                }
                            }
                        }
                    ]
                },
                'reverse': {
                    'name': 'child',
                    'typeId': '2'
                },
                'singular': 'mmap',
                'typeId': '1'
            }
        }
    }
}

snapshots['test_schema_success[update] 2'] = {
    'data': {
        'allProductRelationTypes': {
            'edges': [
                {
                    'node': {
                        'indefArticle': 'an',
                        'name': 'parent',
                        'plural': 'mmaps',
                        'relations': {
                            'edges': [
                                {
                                    'node': {
                                        'other': {
                                            'name': 'map1',
                                            'productId': '1'
                                        },
                                        'self_': {
                                            'name': 'beam1',
                                            'productId': '4'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'other': {
                                            'name': 'map1',
                                            'productId': '1'
                                        },
                                        'self_': {
                                            'name': 'beam2',
                                            'productId': '5'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'other': {
                                            'name': 'beam1',
                                            'productId': '4'
                                        },
                                        'self_': {
                                            'name': 'beam2',
                                            'productId': '5'
                                        }
                                    }
                                }
                            ]
                        },
                        'reverse': {
                            'name': 'child',
                            'typeId': '2'
                        },
                        'singular': 'mmap',
                        'typeId': '1'
                    }
                },
                {
                    'node': {
                        'indefArticle': 'a',
                        'name': 'child',
                        'plural': 'children',
                        'relations': {
                            'edges': [
                                {
                                    'node': {
                                        'other': {
                                            'name': 'beam1',
                                            'productId': '4'
                                        },
                                        'self_': {
                                            'name': 'map1',
                                            'productId': '1'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'other': {
                                            'name': 'beam2',
                                            'productId': '5'
                                        },
                                        'self_': {
                                            'name': 'map1',
                                            'productId': '1'
                                        }
                                    }
                                },
                                {
                                    'node': {
                                        'other': {
                                            'name': 'beam2',
                                            'productId': '5'
                                        },
                                        'self_': {
                                            'name': 'beam1',
                                            'productId': '4'
                                        }
                                    }
                                }
                            ]
                        },
                        'reverse': {
                            'name': 'parent',
                            'typeId': '1'
                        },
                        'singular': 'child',
                        'typeId': '2'
                    }
                },
                {
                    'node': {
                        'indefArticle': 'a',
                        'name': 'plaintiff',
                        'plural': 'plaintiffs',
                        'relations': {
                            'edges': [
                            ]
                        },
                        'reverse': {
                            'name': 'defendant',
                            'typeId': '4'
                        },
                        'singular': 'plaintiff',
                        'typeId': '3'
                    }
                },
                {
                    'node': {
                        'indefArticle': 'a',
                        'name': 'defendant',
                        'plural': 'defendants',
                        'relations': {
                            'edges': [
                            ]
                        },
                        'reverse': {
                            'name': 'plaintiff',
                            'typeId': '3'
                        },
                        'singular': 'defendant',
                        'typeId': '4'
                    }
                }
            ]
        }
    }
}



"""copy attributes

Revision ID: 35e6ddccd22a
Revises: 1583f2de9f39
Create Date: 2021-09-23 20:35:53.480108

"""
from alembic import op

from sqlalchemy.orm.session import Session

from sqlalchemy import MetaData
from flask_sqlalchemy import SQLAlchemy

# revision identifiers, used by Alembic.
revision = "35e6ddccd22a"
down_revision = "1583f2de9f39"
branch_labels = None
depends_on = None

#
convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    # "ck": "ck_%(table_name)s_%(constraint_name)s",
    "ck": "ck_%(table_name)s_%(column_0_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}
metadata = MetaData(naming_convention=convention)

sa = SQLAlchemy(metadata=metadata)

#
import enum  # noqa: E402
import datetime  # noqa: E402

from sqlalchemy.orm import declared_attr  # noqa: E402
from sqlalchemy.orm import declarative_mixin  # noqa: E402


class ProductType(sa.Model):
    __tablename__ = "product_types"
    type_id = sa.Column(sa.Integer(), primary_key=True)
    name = sa.Column(sa.Text(), nullable=False, unique=True, index=True)
    order = sa.Column(sa.Integer())
    indef_article = sa.Column(sa.Text())
    singular = sa.Column(sa.Text())
    plural = sa.Column(sa.Text())
    icon = sa.Column(sa.Text())

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name!r}>"


class Product(sa.Model):
    # TODO: rename the class name to be more generic, e.g., "Entry"
    # TODO: use singular for the table name
    __tablename__ = "products"
    product_id = sa.Column(sa.Integer(), primary_key=True)
    type_id = sa.Column(sa.ForeignKey("product_types.type_id"), nullable=False)
    type_ = sa.relationship("ProductType", backref=sa.backref("products"))
    name = sa.Column(sa.Text(), nullable=False)
    contact = sa.Column(sa.Text())  # TODO: move to attribute model
    date_produced = sa.Column(sa.Date())  # TODO: move to attribute model
    produced_by = sa.Column(sa.Text())  # TODO: move to attribute model
    time_posted = sa.Column(
        sa.DateTime(), default=lambda: datetime.datetime.now()
    )
    posted_by = sa.Column(sa.Text())
    posting_git_hub_user_id = sa.Column(sa.ForeignKey("github_users.user_id"))
    posting_git_hub_user = sa.relationship(
        "GitHubUser",
        foreign_keys=[posting_git_hub_user_id],
        backref=sa.backref("posted_products", cascade="all"),
    )
    time_updated = sa.Column(sa.DateTime())
    updated_by = sa.Column(sa.Text())
    updating_git_hub_user_id = sa.Column(sa.ForeignKey("github_users.user_id"))
    updating_git_hub_user = sa.relationship(
        "GitHubUser",
        foreign_keys=[updating_git_hub_user_id],
        backref=sa.backref("updated_products", cascade="all"),
    )
    note = sa.Column(sa.Text())
    __table_args__ = (
        sa.UniqueConstraint("type_id", "name", name="_type_id_name"),
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name!r}>"


class GitHubUser(sa.Model):
    __tablename__ = "github_users"
    user_id = sa.Column(sa.Integer(), primary_key=True)
    git_hub_id = sa.Column(sa.Text(), unique=True, nullable=False)
    login = sa.Column(sa.Text(), unique=True, nullable=False)
    name = sa.Column(sa.Text())
    avatar_url = sa.Column(sa.Text())
    url = sa.Column(sa.Text())

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.login!r}>"


class FieldType(enum.Enum):
    # SQLAlchemy generic types: https://docs.sqlalchemy.org/en/14/core/type_basics.html#generic-types
    UnicodeText = 1
    Boolean = 2
    Integer = 3
    Float = 4
    Date = 5
    DateTime = 6
    Time = 7


class Field(sa.Model):
    __tablename__ = "field"
    field_id = sa.Column(sa.Integer(), primary_key=True)
    name = sa.Column(sa.UnicodeText(), nullable=False)
    type_ = sa.Column(sa.Enum(FieldType), nullable=False)

    def __repr__(self):
        type_name = self.type_.name if self.type_ else self.type_
        return f"<{self.__class__.__name__} {self.name!r} {type_name!r}>"


class TypeFieldAssociation(sa.Model):
    __tablename__ = "type_field_association"
    type_id = sa.Column(
        #
        sa.Integer(),
        sa.ForeignKey("product_types.type_id"),
        primary_key=True,
    )
    field_id = sa.Column(
        #
        sa.Integer(),
        sa.ForeignKey("field.field_id"),
        primary_key=True,
    )
    type_ = sa.relationship(
        "ProductType",
        backref=sa.backref("fields", cascade="all, delete-orphan"),
    )
    field = sa.relationship(
        "Field",
        backref=sa.backref("entry_types", cascade="all, delete-orphan"),
    )

    def __repr__(self):
        type_name = self.type_.name if self.type_ else self.type_
        field_name = self.field.name if self.field else self.field
        return f"<{self.__class__.__name__} {type_name!r} {field_name!r}>"


@declarative_mixin
class AttributeBase:
    attribute_id = sa.Column(sa.Integer(), primary_key=True)
    name = sa.Column(sa.Text())  # TODO: Remove

    @declared_attr
    def product_id(self):
        return sa.Column(sa.Integer(), sa.ForeignKey("products.product_id"), nullable=False)  # fmt: skip

    @declared_attr
    def product(self):
        return sa.relationship(
            "Product",
            backref=sa.backref(self.backref_column, cascade="all, delete-orphan"),  # fmt: skip
        )

    @declared_attr
    def field_id(self):
        return sa.Column(sa.Integer(), sa.ForeignKey("field.field_id"), nullable=True)  # fmt: skip
        # TODO: Make nullable False

    @declared_attr
    def field(self):
        return sa.relationship(
            "Field",
            backref=sa.backref(self.backref_column, cascade="all, delete-orphan"),  # fmt: skip
        )

    def __repr__(self):
        field_name = self.field.name if self.field else self.field
        return f"<{self.__class__.__name__} {field_name!r}: {self.value!r}>"


class AttributeUnicodeText(AttributeBase, sa.Model):
    __tablename__ = "attribute_unicode_text"
    backref_column = "attributes_unicode_text"
    value = sa.Column(sa.UnicodeText())


class AttributeBoolean(AttributeBase, sa.Model):
    __tablename__ = "attribute_boolean"
    backref_column = "attributes_boolean"
    value = sa.Column(sa.Boolean())


class AttributeInteger(AttributeBase, sa.Model):
    __tablename__ = "attribute_integer"
    backref_column = "attributes_integer"
    value = sa.Column(sa.Integer())


class AttributeFloat(AttributeBase, sa.Model):
    __tablename__ = "attribute_float"
    backref_column = "attributes_float"
    value = sa.Column(sa.Float())


class AttributeDate(AttributeBase, sa.Model):
    __tablename__ = "attribute_date"
    backref_column = "attributes_date"
    value = sa.Column(sa.Date())


class AttributeDateTime(AttributeBase, sa.Model):
    __tablename__ = "attribute_date_time"
    backref_column = "attributes_date_time"
    value = sa.Column(sa.DateTime())


class AttributeTime(AttributeBase, sa.Model):
    __tablename__ = "attribute_time"
    backref_column = "attributes_time"
    value = sa.Column(sa.Time())


def upgrade():
    session = Session(bind=op.get_bind())

    AttributeClasses = [AttributeUnicodeText, AttributeDate, AttributeDateTime]

    for product in session.query(Product):
        field_dict = {f.field.name: f.field for f in product.type_.fields}
        for AttributeClass in AttributeClasses:
            attrs = getattr(product, AttributeClass.backref_column)
            for attr in attrs:
                attr.field = field_dict[attr.name]
    session.commit()


def downgrade():
    pass



import graphene

from . import type_

from ... import ops


class CreateLogInput(graphene.InputObjectType):
    """Input to createLog()"""

    id_ = graphene.Int()
    level = graphene.String(description="The log level")
    message = graphene.String(description="The message")


class CreateLog(graphene.Mutation):
    """Create a log"""

    class Arguments:
        input = CreateLogInput(required=True)

    ok = graphene.Boolean()
    log = graphene.Field(lambda: type_.Log)

    def mutate(root, info, input):
        model = ops.create_log(**input)
        ops.commit()
        ok = True
        return CreateLog(log=model, ok=ok)


class DeleteLog(graphene.Mutation):
    """Delete a log"""

    class Arguments:
        id_ = graphene.Int(description="The id of the log")

    ok = graphene.Boolean()

    def mutate(root, info, id_):
        ops.delete_log(id_=id_)
        ops.commit()
        ok = True
        return DeleteLog(ok=ok)



"""Operations on the SQLAlchemy ORM models

This package contains operations on the SQLAlchemy ORM models, e.g.,
instantiation.

"""

from .misc import commit  # noqa: F401
from .product import (  # noqa: F401
    create_product,
    update_product,
    delete_product,
)
from .product_type import (  # noqa: F401
    create_product_type,
    update_product_type,
    delete_product_type,
)
from .product_relation import (  # noqa: F401
    create_product_relation,
    delete_product_relation,
)
from .product_relation_type import (  # noqa: F401
    create_product_relation_type,
    update_product_relation_type,
    delete_product_relation_type,
)
from .product_file_path import (  # noqa: F401
    create_product_file_path,
    update_product_file_path,
    delete_product_file_path,
)
from .field import (  # noqa: F401
    create_field,
    update_field,
    delete_field,
    FieldType,
)
from .log import (  # noqa: F401
    create_log,
    delete_log,
)



import graphene

from ...models import Log as LogModel

from ..filter_ import PFilterableConnectionField
from . import type_


##__________________________________________________________________||
all_logs_field = PFilterableConnectionField(type_.Log.connection)


##__________________________________________________________________||
def resolve_log(parent, info, **kwargs):
    filter = [getattr(LogModel, k) == v for k, v in kwargs.items()]
    return type_.Log.get_query(info).filter(*filter).one_or_none()


log_field = graphene.Field(type_.Log, id_=graphene.Int(), resolver=resolve_log)

##__________________________________________________________________||



"""add tables

Revision ID: e077a31ad00b
Revises: 4398f463d243
Create Date: 2021-09-22 14:28:11.609615

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'e077a31ad00b'
down_revision = '4398f463d243'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('field',
    sa.Column('field_id', sa.Integer(), nullable=False),
    sa.Column('name', sa.UnicodeText(), nullable=False),
    sa.Column('type_', sa.Enum('UnicodeText', 'Boolean', 'Int', 'Float', 'Date', 'DateTime', 'Time', name='fieldtype'), nullable=False),
    sa.PrimaryKeyConstraint('field_id', name=op.f('pk_field'))
    )
    op.create_table('type_field_association',
    sa.Column('type_id', sa.Integer(), nullable=False),
    sa.Column('field_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['field_id'], ['field.field_id'], name=op.f('fk_type_field_association_field_id_field')),
    sa.ForeignKeyConstraint(['type_id'], ['product_types.type_id'], name=op.f('fk_type_field_association_type_id_product_types')),
    sa.PrimaryKeyConstraint('type_id', 'field_id', name=op.f('pk_type_field_association'))
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('type_field_association')
    op.drop_table('field')
    # ### end Alembic commands ###



# fmt: off
from .mutation_create_product import MUTATION_CREATE_PRODUCT  # noqa: F401
from .mutation_delete_product import MUTATION_DELETE_PRODUCT  # noqa: F401
from .mutation_update_product import MUTATION_UPDATE_PRODUCT  # noqa: F401
from .mutation_create_product_relation_types import MUTATION_CREATE_PRODUCT_RELATION_TYPES  # noqa: F401
from .mutation_update_product_relation_type import MUTATION_UPDATE_PRODUCT_RELATION_TYPE  # noqa: F401
from .mutation_delete_product_relation_types import MUTATION_DELETE_PRODUCT_RELATION_TYPES  # noqa: F401
from .mutation_create_product_relation import MUTATION_CREATE_PRODUCT_RELATION  # noqa: F401
from .mutation_delete_product_relation import MUTATION_DELETE_PRODUCT_RELATION  # noqa: F401
from .mutation_create_product_type import MUTATION_CREATE_PRODUCT_TYPE  # noqa: F401
from .mutation_update_product_type import MUTATION_UPDATE_PRODUCT_TYPE  # noqa: F401
from .mutation_delete_product_type import MUTATION_DELETE_PRODUCT_TYPE  # noqa: F401
from .mutation_create_field import MUTATION_CREATE_FIELD  # noqa: F401
from .mutation_update_field import MUTATION_UPDATE_FIELD  # noqa: F401
from .mutation_delete_field import MUTATION_DELETE_FIELD  # noqa: F401



from ..db.sa import sa
from ..models import Log


def create_log(**kwargs):
    model = Log(**kwargs)
    sa.session.add(model)
    return model


def delete_log(id_):
    model = Log.query.filter_by(id_=id_).one()
    sa.session.delete(model)
    return



from ..fragments import FRAGMENT_PRODUCT_RELATION

MUTATION_CREATE_PRODUCT_RELATION = '''
mutation CreateProductRelation($input: CreateProductRelationInput!) {
  createProductRelation(input: $input) {
    ok
    productRelation {
      ...fragmentProductRelation
    }
  }
}
''' + FRAGMENT_PRODUCT_RELATION



from ..fragments import FRAGMENT_PRODUCT_RELATION_TYPE

MUTATION_UPDATE_PRODUCT_RELATION_TYPE = '''
mutation UpdateProductRelationType(
  $typeId: Int!
  $input: UpdateProductRelationTypeInput!
) {
  updateProductRelationType(typeId: $typeId, input: $input) {
    ok
    productRelationType {
      ...fragmentProductRelationType
    }
  }
}
''' + FRAGMENT_PRODUCT_RELATION_TYPE
